/**
 * Copyright (c) 2010-2012, Mark Czotter, Istvan Rath and Daniel Varro
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors:
 *   Mark Czotter - initial API and implementation
 */
package org.eclipse.viatra.query.patternlanguage.emf.jvmmodel;

import com.google.common.base.Objects;
import com.google.common.collect.Iterables;
import com.google.common.collect.LinkedHashMultimap;
import com.google.common.collect.Sets;
import com.google.inject.Inject;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Set;
import org.eclipse.emf.common.util.EList;
import org.eclipse.viatra.query.patternlanguage.emf.jvmmodel.BodyCodeGenerator;
import org.eclipse.viatra.query.patternlanguage.emf.jvmmodel.JavadocInferrer;
import org.eclipse.viatra.query.patternlanguage.emf.util.EMFJvmTypesBuilder;
import org.eclipse.viatra.query.patternlanguage.emf.util.EMFPatternLanguageGeneratorConfig;
import org.eclipse.viatra.query.patternlanguage.emf.util.EMFPatternLanguageJvmModelInferrerUtil;
import org.eclipse.viatra.query.patternlanguage.emf.util.IErrorFeedback;
import org.eclipse.viatra.query.patternlanguage.emf.validation.EMFIssueCodes;
import org.eclipse.viatra.query.patternlanguage.helper.CorePatternLanguageHelper;
import org.eclipse.viatra.query.patternlanguage.patternLanguage.Annotation;
import org.eclipse.viatra.query.patternlanguage.patternLanguage.ExecutionType;
import org.eclipse.viatra.query.patternlanguage.patternLanguage.Modifiers;
import org.eclipse.viatra.query.patternlanguage.patternLanguage.Parameter;
import org.eclipse.viatra.query.patternlanguage.patternLanguage.ParameterDirection;
import org.eclipse.viatra.query.patternlanguage.patternLanguage.Pattern;
import org.eclipse.viatra.query.patternlanguage.patternLanguage.PatternBody;
import org.eclipse.viatra.query.patternlanguage.patternLanguage.Type;
import org.eclipse.viatra.query.patternlanguage.patternLanguage.Variable;
import org.eclipse.viatra.query.patternlanguage.typing.ITypeInferrer;
import org.eclipse.viatra.query.patternlanguage.typing.ITypeSystem;
import org.eclipse.viatra.query.runtime.api.IQuerySpecification;
import org.eclipse.viatra.query.runtime.api.ViatraQueryEngine;
import org.eclipse.viatra.query.runtime.api.impl.BaseGeneratedEMFPQuery;
import org.eclipse.viatra.query.runtime.api.impl.BaseGeneratedEMFQuerySpecification;
import org.eclipse.viatra.query.runtime.api.impl.BaseGeneratedEMFQuerySpecificationWithGenericMatcher;
import org.eclipse.viatra.query.runtime.exception.ViatraQueryException;
import org.eclipse.viatra.query.runtime.localsearch.matcher.integration.LocalSearchBackendFactory;
import org.eclipse.viatra.query.runtime.matchers.backend.QueryEvaluationHint;
import org.eclipse.viatra.query.runtime.matchers.context.IInputKey;
import org.eclipse.viatra.query.runtime.matchers.psystem.PBody;
import org.eclipse.viatra.query.runtime.matchers.psystem.annotations.PAnnotation;
import org.eclipse.viatra.query.runtime.matchers.psystem.annotations.ParameterReference;
import org.eclipse.viatra.query.runtime.matchers.psystem.queries.PParameter;
import org.eclipse.viatra.query.runtime.matchers.psystem.queries.PParameterDirection;
import org.eclipse.viatra.query.runtime.matchers.psystem.queries.PProblem;
import org.eclipse.viatra.query.runtime.matchers.psystem.queries.QueryInitializationException;
import org.eclipse.viatra.query.runtime.rete.matcher.ReteBackendFactory;
import org.eclipse.xtend2.lib.StringConcatenation;
import org.eclipse.xtend2.lib.StringConcatenationClient;
import org.eclipse.xtext.common.types.JvmAnnotationReference;
import org.eclipse.xtext.common.types.JvmConstructor;
import org.eclipse.xtext.common.types.JvmDeclaredType;
import org.eclipse.xtext.common.types.JvmField;
import org.eclipse.xtext.common.types.JvmFormalParameter;
import org.eclipse.xtext.common.types.JvmGenericType;
import org.eclipse.xtext.common.types.JvmMember;
import org.eclipse.xtext.common.types.JvmOperation;
import org.eclipse.xtext.common.types.JvmType;
import org.eclipse.xtext.common.types.JvmTypeReference;
import org.eclipse.xtext.common.types.JvmUnknownTypeReference;
import org.eclipse.xtext.common.types.JvmVisibility;
import org.eclipse.xtext.diagnostics.Severity;
import org.eclipse.xtext.serializer.impl.Serializer;
import org.eclipse.xtext.xbase.XExpression;
import org.eclipse.xtext.xbase.jvmmodel.JvmAnnotationReferenceBuilder;
import org.eclipse.xtext.xbase.jvmmodel.JvmTypeReferenceBuilder;
import org.eclipse.xtext.xbase.lib.Exceptions;
import org.eclipse.xtext.xbase.lib.Extension;
import org.eclipse.xtext.xbase.lib.Functions.Function1;
import org.eclipse.xtext.xbase.lib.IterableExtensions;
import org.eclipse.xtext.xbase.lib.ListExtensions;
import org.eclipse.xtext.xbase.lib.Procedures.Procedure1;

/**
 * {@link IQuerySpecification} implementation inferrer.
 * 
 * @author Mark Czotter
 * @noreference
 */
@SuppressWarnings("all")
public class PatternQuerySpecificationClassInferrer {
  @Inject
  @Extension
  private EMFJvmTypesBuilder _eMFJvmTypesBuilder;
  
  @Inject
  @Extension
  private EMFPatternLanguageJvmModelInferrerUtil util;
  
  @Inject
  @Extension
  private JavadocInferrer _javadocInferrer;
  
  @Inject
  @Extension
  private ITypeInferrer _iTypeInferrer;
  
  @Inject
  private ITypeSystem typeSystem;
  
  @Inject
  private IErrorFeedback feedback;
  
  @Inject
  private Serializer serializer;
  
  @Extension
  private JvmTypeReferenceBuilder builder;
  
  @Extension
  private JvmAnnotationReferenceBuilder annBuilder;
  
  /**
   * Infers the {@link IQuerySpecification} implementation class from {@link Pattern}.
   */
  public JvmDeclaredType inferQuerySpecificationClass(final Pattern pattern, final boolean isPrelinkingPhase, final String querySpecificationPackageName, final JvmType matcherClass, final JvmTypeReferenceBuilder builder, final JvmAnnotationReferenceBuilder annBuilder, final EMFPatternLanguageGeneratorConfig config) {
    this.builder = builder;
    this.annBuilder = annBuilder;
    EMFPatternLanguageGeneratorConfig.MatcherGenerationStrategy _matcherGenerationStrategy = config.getMatcherGenerationStrategy();
    String _querySpecificationClassName = this.util.querySpecificationClassName(pattern, _matcherGenerationStrategy);
    final Procedure1<JvmGenericType> _function = new Procedure1<JvmGenericType>() {
      @Override
      public void apply(final JvmGenericType it) {
        it.setPackageName(querySpecificationPackageName);
        CharSequence _javadocQuerySpecificationClass = PatternQuerySpecificationClassInferrer.this._javadocInferrer.javadocQuerySpecificationClass(pattern);
        String _string = _javadocQuerySpecificationClass.toString();
        PatternQuerySpecificationClassInferrer.this._eMFJvmTypesBuilder.setDocumentation(it, _string);
        it.setFinal(true);
        boolean _and = false;
        boolean _isPublic = PatternQuerySpecificationClassInferrer.this.util.isPublic(pattern);
        if (!_isPublic) {
          _and = false;
        } else {
          EMFPatternLanguageGeneratorConfig.MatcherGenerationStrategy _matcherGenerationStrategy = config.getMatcherGenerationStrategy();
          boolean _tripleNotEquals = (_matcherGenerationStrategy != EMFPatternLanguageGeneratorConfig.MatcherGenerationStrategy.USE_GENERIC);
          _and = _tripleNotEquals;
        }
        if (_and) {
          EList<JvmTypeReference> _superTypes = it.getSuperTypes();
          JvmTypeReference _typeRef = PatternQuerySpecificationClassInferrer.this.builder.typeRef(matcherClass);
          JvmTypeReference _typeRef_1 = PatternQuerySpecificationClassInferrer.this.builder.typeRef(BaseGeneratedEMFQuerySpecification.class, _typeRef);
          PatternQuerySpecificationClassInferrer.this._eMFJvmTypesBuilder.<JvmTypeReference>operator_add(_superTypes, _typeRef_1);
        } else {
          EList<JvmTypeReference> _superTypes_1 = it.getSuperTypes();
          JvmTypeReference _typeRef_2 = PatternQuerySpecificationClassInferrer.this.builder.typeRef(BaseGeneratedEMFQuerySpecificationWithGenericMatcher.class);
          PatternQuerySpecificationClassInferrer.this._eMFJvmTypesBuilder.<JvmTypeReference>operator_add(_superTypes_1, _typeRef_2);
        }
        String _fileComment = PatternQuerySpecificationClassInferrer.this.util.getFileComment(pattern);
        PatternQuerySpecificationClassInferrer.this._eMFJvmTypesBuilder.setFileHeader(it, _fileComment);
      }
    };
    final JvmGenericType querySpecificationClass = this._eMFJvmTypesBuilder.toClass(pattern, _querySpecificationClassName, _function);
    return querySpecificationClass;
  }
  
  public void initializeSpecification(final JvmDeclaredType querySpecificationClass, final Pattern pattern, final JvmType matcherClass, final JvmType matchClass, final EMFPatternLanguageGeneratorConfig config) {
    try {
      boolean _and = false;
      boolean _isPublic = this.util.isPublic(pattern);
      if (!_isPublic) {
        _and = false;
      } else {
        EMFPatternLanguageGeneratorConfig.MatcherGenerationStrategy _matcherGenerationStrategy = config.getMatcherGenerationStrategy();
        boolean _tripleNotEquals = (_matcherGenerationStrategy != EMFPatternLanguageGeneratorConfig.MatcherGenerationStrategy.USE_GENERIC);
        _and = _tripleNotEquals;
      }
      final boolean withPatternSpecificMatcher = _and;
      this.inferQuerySpecificationMethods(querySpecificationClass, pattern, matcherClass, matchClass, withPatternSpecificMatcher);
      this.inferQuerySpecificationInnerClasses(querySpecificationClass, pattern, withPatternSpecificMatcher);
      this.inferExpressions(querySpecificationClass, pattern);
    } catch (final Throwable _t) {
      if (_t instanceof IllegalStateException) {
        final IllegalStateException ex = (IllegalStateException)_t;
        String _message = ex.getMessage();
        this.feedback.reportError(pattern, _message, EMFIssueCodes.OTHER_ISSUE, Severity.ERROR, 
          IErrorFeedback.JVMINFERENCE_ERROR_TYPE);
      } else {
        throw Exceptions.sneakyThrow(_t);
      }
    }
  }
  
  /**
   * Infers methods for QuerySpecification class based on the input 'pattern'.
   */
  public boolean inferQuerySpecificationMethods(final JvmDeclaredType querySpecificationClass, final Pattern pattern, final JvmType matcherClass, final JvmType matchClass, final boolean withPatternSpecificMatcher) {
    boolean _xblockexpression = false;
    {
      EList<JvmMember> _members = querySpecificationClass.getMembers();
      final Procedure1<JvmConstructor> _function = new Procedure1<JvmConstructor>() {
        @Override
        public void apply(final JvmConstructor it) {
          it.setVisibility(JvmVisibility.PRIVATE);
          StringConcatenationClient _client = new StringConcatenationClient() {
            @Override
            protected void appendTo(StringConcatenationClient.TargetStringConcatenation _builder) {
              _builder.append("super(");
              String _querySpecificationPQueryClassName = PatternQuerySpecificationClassInferrer.this.util.querySpecificationPQueryClassName(pattern);
              _builder.append(_querySpecificationPQueryClassName, "");
              _builder.append(".INSTANCE);");
              _builder.newLineIfNotEmpty();
            }
          };
          PatternQuerySpecificationClassInferrer.this._eMFJvmTypesBuilder.setBody(it, _client);
        }
      };
      JvmConstructor _constructor = this._eMFJvmTypesBuilder.toConstructor(pattern, _function);
      this._eMFJvmTypesBuilder.<JvmConstructor>operator_add(_members, _constructor);
      EList<JvmMember> _members_1 = querySpecificationClass.getMembers();
      JvmTypeReference _typeRef = this.builder.typeRef(querySpecificationClass);
      final Procedure1<JvmOperation> _function_1 = new Procedure1<JvmOperation>() {
        @Override
        public void apply(final JvmOperation it) {
          it.setVisibility(JvmVisibility.PUBLIC);
          it.setStatic(true);
          EList<JvmTypeReference> _exceptions = it.getExceptions();
          JvmTypeReference _typeRef = PatternQuerySpecificationClassInferrer.this.builder.typeRef(ViatraQueryException.class);
          PatternQuerySpecificationClassInferrer.this._eMFJvmTypesBuilder.<JvmTypeReference>operator_add(_exceptions, _typeRef);
          CharSequence _javadocQuerySpecificationInstanceMethod = PatternQuerySpecificationClassInferrer.this._javadocInferrer.javadocQuerySpecificationInstanceMethod(pattern);
          String _string = _javadocQuerySpecificationInstanceMethod.toString();
          PatternQuerySpecificationClassInferrer.this._eMFJvmTypesBuilder.setDocumentation(it, _string);
          StringConcatenationClient _client = new StringConcatenationClient() {
            @Override
            protected void appendTo(StringConcatenationClient.TargetStringConcatenation _builder) {
              _builder.append("try{");
              _builder.newLine();
              _builder.append("    ");
              _builder.append("return ");
              String _querySpecificationHolderClassName = PatternQuerySpecificationClassInferrer.this.util.querySpecificationHolderClassName(pattern);
              _builder.append(_querySpecificationHolderClassName, "    ");
              _builder.append(".INSTANCE;");
              _builder.newLineIfNotEmpty();
              _builder.append("} catch (");
              _builder.append(ExceptionInInitializerError.class, "");
              _builder.append(" err) {");
              _builder.newLineIfNotEmpty();
              _builder.append("    ");
              _builder.append("throw processInitializerError(err);");
              _builder.newLine();
              _builder.append("}");
              _builder.newLine();
            }
          };
          PatternQuerySpecificationClassInferrer.this._eMFJvmTypesBuilder.setBody(it, _client);
        }
      };
      JvmOperation _method = this._eMFJvmTypesBuilder.toMethod(pattern, "instance", _typeRef, _function_1);
      this._eMFJvmTypesBuilder.<JvmOperation>operator_add(_members_1, _method);
      boolean _xifexpression = false;
      if (withPatternSpecificMatcher) {
        boolean _xblockexpression_1 = false;
        {
          EList<JvmMember> _members_2 = querySpecificationClass.getMembers();
          JvmTypeReference _typeRef_1 = this.builder.typeRef(matcherClass);
          final Procedure1<JvmOperation> _function_2 = new Procedure1<JvmOperation>() {
            @Override
            public void apply(final JvmOperation it) {
              it.setVisibility(JvmVisibility.PROTECTED);
              EList<JvmAnnotationReference> _annotations = it.getAnnotations();
              JvmAnnotationReference _annotationRef = PatternQuerySpecificationClassInferrer.this.annBuilder.annotationRef(Override.class);
              PatternQuerySpecificationClassInferrer.this._eMFJvmTypesBuilder.<JvmAnnotationReference>operator_add(_annotations, _annotationRef);
              EList<JvmFormalParameter> _parameters = it.getParameters();
              JvmTypeReference _typeRef = PatternQuerySpecificationClassInferrer.this.builder.typeRef(ViatraQueryEngine.class);
              JvmFormalParameter _parameter = PatternQuerySpecificationClassInferrer.this._eMFJvmTypesBuilder.toParameter(pattern, "engine", _typeRef);
              PatternQuerySpecificationClassInferrer.this._eMFJvmTypesBuilder.<JvmFormalParameter>operator_add(_parameters, _parameter);
              EList<JvmTypeReference> _exceptions = it.getExceptions();
              JvmTypeReference _typeRef_1 = PatternQuerySpecificationClassInferrer.this.builder.typeRef(ViatraQueryException.class);
              PatternQuerySpecificationClassInferrer.this._eMFJvmTypesBuilder.<JvmTypeReference>operator_add(_exceptions, _typeRef_1);
              StringConcatenationClient _client = new StringConcatenationClient() {
                @Override
                protected void appendTo(StringConcatenationClient.TargetStringConcatenation _builder) {
                  _builder.append("return ");
                  _builder.append(matcherClass, "");
                  _builder.append(".on(engine);");
                }
              };
              PatternQuerySpecificationClassInferrer.this._eMFJvmTypesBuilder.setBody(it, _client);
            }
          };
          JvmOperation _method_1 = this._eMFJvmTypesBuilder.toMethod(pattern, "instantiate", _typeRef_1, _function_2);
          this._eMFJvmTypesBuilder.<JvmOperation>operator_add(_members_2, _method_1);
          EList<JvmMember> _members_3 = querySpecificationClass.getMembers();
          JvmTypeReference _typeRef_2 = this.builder.typeRef(matcherClass);
          final Procedure1<JvmOperation> _function_3 = new Procedure1<JvmOperation>() {
            @Override
            public void apply(final JvmOperation it) {
              it.setVisibility(JvmVisibility.PUBLIC);
              EList<JvmAnnotationReference> _annotations = it.getAnnotations();
              JvmAnnotationReference _annotationRef = PatternQuerySpecificationClassInferrer.this.annBuilder.annotationRef(Override.class);
              PatternQuerySpecificationClassInferrer.this._eMFJvmTypesBuilder.<JvmAnnotationReference>operator_add(_annotations, _annotationRef);
              EList<JvmTypeReference> _exceptions = it.getExceptions();
              JvmTypeReference _typeRef = PatternQuerySpecificationClassInferrer.this.builder.typeRef(ViatraQueryException.class);
              PatternQuerySpecificationClassInferrer.this._eMFJvmTypesBuilder.<JvmTypeReference>operator_add(_exceptions, _typeRef);
              StringConcatenationClient _client = new StringConcatenationClient() {
                @Override
                protected void appendTo(StringConcatenationClient.TargetStringConcatenation _builder) {
                  _builder.append("return ");
                  _builder.append(matcherClass, "");
                  _builder.append(".create();");
                }
              };
              PatternQuerySpecificationClassInferrer.this._eMFJvmTypesBuilder.setBody(it, _client);
            }
          };
          JvmOperation _method_2 = this._eMFJvmTypesBuilder.toMethod(pattern, "instantiate", _typeRef_2, _function_3);
          this._eMFJvmTypesBuilder.<JvmOperation>operator_add(_members_3, _method_2);
          EList<JvmMember> _members_4 = querySpecificationClass.getMembers();
          JvmTypeReference _typeRef_3 = this.builder.typeRef(matchClass);
          final Procedure1<JvmOperation> _function_4 = new Procedure1<JvmOperation>() {
            @Override
            public void apply(final JvmOperation it) {
              it.setVisibility(JvmVisibility.PUBLIC);
              EList<JvmAnnotationReference> _annotations = it.getAnnotations();
              JvmAnnotationReference _annotationRef = PatternQuerySpecificationClassInferrer.this.annBuilder.annotationRef(Override.class);
              PatternQuerySpecificationClassInferrer.this._eMFJvmTypesBuilder.<JvmAnnotationReference>operator_add(_annotations, _annotationRef);
              StringConcatenationClient _client = new StringConcatenationClient() {
                @Override
                protected void appendTo(StringConcatenationClient.TargetStringConcatenation _builder) {
                  _builder.append("return ");
                  _builder.append(matchClass, "");
                  _builder.append(".newEmptyMatch();");
                }
              };
              PatternQuerySpecificationClassInferrer.this._eMFJvmTypesBuilder.setBody(it, _client);
            }
          };
          JvmOperation _method_3 = this._eMFJvmTypesBuilder.toMethod(pattern, "newEmptyMatch", _typeRef_3, _function_4);
          this._eMFJvmTypesBuilder.<JvmOperation>operator_add(_members_4, _method_3);
          EList<JvmMember> _members_5 = querySpecificationClass.getMembers();
          JvmTypeReference _typeRef_4 = this.builder.typeRef(matchClass);
          final Procedure1<JvmOperation> _function_5 = new Procedure1<JvmOperation>() {
            @Override
            public void apply(final JvmOperation it) {
              it.setVisibility(JvmVisibility.PUBLIC);
              EList<JvmAnnotationReference> _annotations = it.getAnnotations();
              JvmAnnotationReference _annotationRef = PatternQuerySpecificationClassInferrer.this.annBuilder.annotationRef(Override.class);
              PatternQuerySpecificationClassInferrer.this._eMFJvmTypesBuilder.<JvmAnnotationReference>operator_add(_annotations, _annotationRef);
              EList<JvmFormalParameter> _parameters = it.getParameters();
              JvmTypeReference _typeRef = PatternQuerySpecificationClassInferrer.this.builder.typeRef(Object.class);
              JvmTypeReference _addArrayTypeDimension = PatternQuerySpecificationClassInferrer.this._eMFJvmTypesBuilder.addArrayTypeDimension(_typeRef);
              JvmFormalParameter _parameter = PatternQuerySpecificationClassInferrer.this._eMFJvmTypesBuilder.toParameter(pattern, "parameters", _addArrayTypeDimension);
              PatternQuerySpecificationClassInferrer.this._eMFJvmTypesBuilder.<JvmFormalParameter>operator_add(_parameters, _parameter);
              it.setVarArgs(true);
              StringConcatenationClient _client = new StringConcatenationClient() {
                @Override
                protected void appendTo(StringConcatenationClient.TargetStringConcatenation _builder) {
                  _builder.append("return ");
                  _builder.append(matchClass, "");
                  _builder.append(".newMatch(");
                  {
                    EList<Variable> _parameters = pattern.getParameters();
                    boolean _hasElements = false;
                    for(final Variable p : _parameters) {
                      if (!_hasElements) {
                        _hasElements = true;
                      } else {
                        _builder.appendImmediate(", ", "");
                      }
                      _builder.append("(");
                      JvmTypeReference _calculateType = PatternQuerySpecificationClassInferrer.this.util.calculateType(p);
                      String _qualifiedName = _calculateType.getQualifiedName();
                      _builder.append(_qualifiedName, "");
                      _builder.append(") parameters[");
                      EList<Variable> _parameters_1 = pattern.getParameters();
                      int _indexOf = _parameters_1.indexOf(p);
                      _builder.append(_indexOf, "");
                      _builder.append("]");
                    }
                  }
                  _builder.append(");");
                }
              };
              PatternQuerySpecificationClassInferrer.this._eMFJvmTypesBuilder.setBody(it, _client);
            }
          };
          JvmOperation _method_4 = this._eMFJvmTypesBuilder.toMethod(pattern, "newMatch", _typeRef_4, _function_5);
          _xblockexpression_1 = this._eMFJvmTypesBuilder.<JvmOperation>operator_add(_members_5, _method_4);
        }
        _xifexpression = _xblockexpression_1;
      }
      _xblockexpression = _xifexpression;
    }
    return _xblockexpression;
  }
  
  public ParameterDirection direction(final Variable variable) {
    ParameterDirection _xblockexpression = null;
    {
      if ((variable instanceof Parameter)) {
        ((Parameter)variable).getDirection();
      }
      _xblockexpression = ParameterDirection.INOUT;
    }
    return _xblockexpression;
  }
  
  public StringConcatenationClient directionLiteral(final Variable variable) {
    StringConcatenationClient _client = new StringConcatenationClient() {
      @Override
      protected void appendTo(StringConcatenationClient.TargetStringConcatenation _builder) {
        _builder.append(PParameterDirection.class, "");
        _builder.append(".");
        ParameterDirection _direction = PatternQuerySpecificationClassInferrer.this.direction(variable);
        String _name = _direction.name();
        _builder.append(_name, "");
      }
    };
    return _client;
  }
  
  public boolean inferPQueryMembers(final JvmDeclaredType pQueryClass, final Pattern pattern) {
    boolean _xblockexpression = false;
    {
      EList<JvmMember> _members = pQueryClass.getMembers();
      JvmTypeReference _typeRef = this.builder.typeRef(pQueryClass);
      final Procedure1<JvmField> _function = new Procedure1<JvmField>() {
        @Override
        public void apply(final JvmField it) {
          it.setFinal(true);
          it.setStatic(true);
          StringConcatenationClient _client = new StringConcatenationClient() {
            @Override
            protected void appendTo(StringConcatenationClient.TargetStringConcatenation _builder) {
              _builder.append("new ");
              String _querySpecificationPQueryClassName = PatternQuerySpecificationClassInferrer.this.util.querySpecificationPQueryClassName(pattern);
              _builder.append(_querySpecificationPQueryClassName, "");
              _builder.append("()");
            }
          };
          PatternQuerySpecificationClassInferrer.this._eMFJvmTypesBuilder.setInitializer(it, _client);
        }
      };
      JvmField _field = this._eMFJvmTypesBuilder.toField(pattern, "INSTANCE", _typeRef, _function);
      this._eMFJvmTypesBuilder.<JvmField>operator_add(_members, _field);
      EList<Variable> _parameters = pattern.getParameters();
      for (final Variable parameter : _parameters) {
        EList<JvmMember> _members_1 = pQueryClass.getMembers();
        String _pParameterName = this.util.getPParameterName(parameter);
        JvmTypeReference _typeRef_1 = this.builder.typeRef(PParameter.class);
        final Procedure1<JvmField> _function_1 = new Procedure1<JvmField>() {
          @Override
          public void apply(final JvmField it) {
            it.setFinal(true);
            it.setVisibility(JvmVisibility.PRIVATE);
            StringConcatenationClient _parameterInstantiation = PatternQuerySpecificationClassInferrer.this.parameterInstantiation(parameter);
            PatternQuerySpecificationClassInferrer.this._eMFJvmTypesBuilder.setInitializer(it, _parameterInstantiation);
          }
        };
        JvmField _field_1 = this._eMFJvmTypesBuilder.toField(pattern, _pParameterName, _typeRef_1, _function_1);
        this._eMFJvmTypesBuilder.<JvmField>operator_add(_members_1, _field_1);
      }
      EList<JvmMember> _members_2 = pQueryClass.getMembers();
      JvmTypeReference _typeRef_2 = this.builder.typeRef(PParameter.class);
      JvmTypeReference _typeRef_3 = this.builder.typeRef(List.class, _typeRef_2);
      final Procedure1<JvmField> _function_2 = new Procedure1<JvmField>() {
        @Override
        public void apply(final JvmField it) {
          it.setFinal(true);
          it.setVisibility(JvmVisibility.PRIVATE);
          StringConcatenationClient _client = new StringConcatenationClient() {
            @Override
            protected void appendTo(StringConcatenationClient.TargetStringConcatenation _builder) {
              _builder.append(Arrays.class, "");
              _builder.append(".asList(");
              {
                EList<Variable> _parameters = pattern.getParameters();
                boolean _hasElements = false;
                for(final Variable param : _parameters) {
                  if (!_hasElements) {
                    _hasElements = true;
                  } else {
                    _builder.appendImmediate(", ", "");
                  }
                  String _pParameterName = PatternQuerySpecificationClassInferrer.this.util.getPParameterName(param);
                  _builder.append(_pParameterName, "");
                }
              }
              _builder.append(")");
            }
          };
          PatternQuerySpecificationClassInferrer.this._eMFJvmTypesBuilder.setInitializer(it, _client);
        }
      };
      JvmField _field_2 = this._eMFJvmTypesBuilder.toField(pattern, "parameters", _typeRef_3, _function_2);
      this._eMFJvmTypesBuilder.<JvmField>operator_add(_members_2, _field_2);
      EList<JvmMember> _members_3 = pQueryClass.getMembers();
      JvmTypeReference _typeRef_4 = this.builder.typeRef(String.class);
      final Procedure1<JvmOperation> _function_3 = new Procedure1<JvmOperation>() {
        @Override
        public void apply(final JvmOperation it) {
          it.setVisibility(JvmVisibility.PUBLIC);
          EList<JvmAnnotationReference> _annotations = it.getAnnotations();
          JvmAnnotationReference _annotationRef = PatternQuerySpecificationClassInferrer.this.annBuilder.annotationRef(Override.class);
          PatternQuerySpecificationClassInferrer.this._eMFJvmTypesBuilder.<JvmAnnotationReference>operator_add(_annotations, _annotationRef);
          StringConcatenationClient _client = new StringConcatenationClient() {
            @Override
            protected void appendTo(StringConcatenationClient.TargetStringConcatenation _builder) {
              _builder.append("return \"");
              String _fullyQualifiedName = CorePatternLanguageHelper.getFullyQualifiedName(pattern);
              _builder.append(_fullyQualifiedName, "");
              _builder.append("\";");
              _builder.newLineIfNotEmpty();
            }
          };
          PatternQuerySpecificationClassInferrer.this._eMFJvmTypesBuilder.setBody(it, _client);
        }
      };
      JvmOperation _method = this._eMFJvmTypesBuilder.toMethod(pattern, "getFullyQualifiedName", _typeRef_4, _function_3);
      this._eMFJvmTypesBuilder.<JvmOperation>operator_add(_members_3, _method);
      EList<JvmMember> _members_4 = pQueryClass.getMembers();
      JvmTypeReference _typeRef_5 = this.builder.typeRef(String.class);
      JvmTypeReference _typeRef_6 = this.builder.typeRef(List.class, _typeRef_5);
      final Procedure1<JvmOperation> _function_4 = new Procedure1<JvmOperation>() {
        @Override
        public void apply(final JvmOperation it) {
          it.setVisibility(JvmVisibility.PUBLIC);
          EList<JvmAnnotationReference> _annotations = it.getAnnotations();
          JvmAnnotationReference _annotationRef = PatternQuerySpecificationClassInferrer.this.annBuilder.annotationRef(Override.class);
          PatternQuerySpecificationClassInferrer.this._eMFJvmTypesBuilder.<JvmAnnotationReference>operator_add(_annotations, _annotationRef);
          StringConcatenationClient _client = new StringConcatenationClient() {
            @Override
            protected void appendTo(StringConcatenationClient.TargetStringConcatenation _builder) {
              _builder.append("return ");
              _builder.append(Arrays.class, "");
              _builder.append(".asList(");
              {
                EList<Variable> _parameters = pattern.getParameters();
                boolean _hasElements = false;
                for(final Variable param : _parameters) {
                  if (!_hasElements) {
                    _hasElements = true;
                  } else {
                    _builder.appendImmediate(",", "");
                  }
                  _builder.append("\"");
                  String _name = param.getName();
                  _builder.append(_name, "");
                  _builder.append("\"");
                }
              }
              _builder.append(");");
            }
          };
          PatternQuerySpecificationClassInferrer.this._eMFJvmTypesBuilder.setBody(it, _client);
        }
      };
      JvmOperation _method_1 = this._eMFJvmTypesBuilder.toMethod(pattern, "getParameterNames", _typeRef_6, _function_4);
      this._eMFJvmTypesBuilder.<JvmOperation>operator_add(_members_4, _method_1);
      EList<JvmMember> _members_5 = pQueryClass.getMembers();
      JvmTypeReference _typeRef_7 = this.builder.typeRef(PParameter.class);
      JvmTypeReference _typeRef_8 = this.builder.typeRef(List.class, _typeRef_7);
      final Procedure1<JvmOperation> _function_5 = new Procedure1<JvmOperation>() {
        @Override
        public void apply(final JvmOperation it) {
          it.setVisibility(JvmVisibility.PUBLIC);
          EList<JvmAnnotationReference> _annotations = it.getAnnotations();
          JvmAnnotationReference _annotationRef = PatternQuerySpecificationClassInferrer.this.annBuilder.annotationRef(Override.class);
          PatternQuerySpecificationClassInferrer.this._eMFJvmTypesBuilder.<JvmAnnotationReference>operator_add(_annotations, _annotationRef);
          StringConcatenationClient _client = new StringConcatenationClient() {
            @Override
            protected void appendTo(StringConcatenationClient.TargetStringConcatenation _builder) {
              _builder.append("return parameters;");
            }
          };
          PatternQuerySpecificationClassInferrer.this._eMFJvmTypesBuilder.setBody(it, _client);
        }
      };
      JvmOperation _method_2 = this._eMFJvmTypesBuilder.toMethod(pattern, "getParameters", _typeRef_8, _function_5);
      this._eMFJvmTypesBuilder.<JvmOperation>operator_add(_members_5, _method_2);
      EList<JvmMember> _members_6 = pQueryClass.getMembers();
      JvmTypeReference _typeRef_9 = this.builder.typeRef(PBody.class);
      JvmTypeReference _typeRef_10 = this.builder.typeRef(Set.class, _typeRef_9);
      final Procedure1<JvmOperation> _function_6 = new Procedure1<JvmOperation>() {
        @Override
        public void apply(final JvmOperation it) {
          it.setVisibility(JvmVisibility.PUBLIC);
          EList<JvmAnnotationReference> _annotations = it.getAnnotations();
          JvmAnnotationReference _annotationRef = PatternQuerySpecificationClassInferrer.this.annBuilder.annotationRef(Override.class);
          PatternQuerySpecificationClassInferrer.this._eMFJvmTypesBuilder.<JvmAnnotationReference>operator_add(_annotations, _annotationRef);
          EList<JvmTypeReference> _exceptions = it.getExceptions();
          JvmTypeReference _typeRef = PatternQuerySpecificationClassInferrer.this.builder.typeRef(QueryInitializationException.class);
          PatternQuerySpecificationClassInferrer.this._eMFJvmTypesBuilder.<JvmTypeReference>operator_add(_exceptions, _typeRef);
          try {
            StringConcatenationClient _client = new StringConcatenationClient() {
              @Override
              protected void appendTo(StringConcatenationClient.TargetStringConcatenation _builder) {
                StringConcatenationClient _inferQueryEvaluationHints = PatternQuerySpecificationClassInferrer.this.inferQueryEvaluationHints(pattern);
                _builder.append(_inferQueryEvaluationHints, "");
                _builder.newLineIfNotEmpty();
                _builder.append(Set.class, "");
                _builder.append("<");
                _builder.append(PBody.class, "");
                _builder.append("> bodies = ");
                _builder.append(Sets.class, "");
                _builder.append(".newLinkedHashSet();");
                _builder.newLineIfNotEmpty();
                {
                  Set<Pattern> _referencedPatterns = CorePatternLanguageHelper.getReferencedPatterns(pattern);
                  final Function1<Pattern, Boolean> _function = new Function1<Pattern, Boolean>() {
                    @Override
                    public Boolean apply(final Pattern it) {
                      return Boolean.valueOf((!Objects.equal(it, pattern)));
                    }
                  };
                  Iterable<Pattern> _filter = IterableExtensions.<Pattern>filter(_referencedPatterns, _function);
                  boolean _isEmpty = IterableExtensions.isEmpty(_filter);
                  if (_isEmpty) {
                    StringConcatenationClient _inferBodies = PatternQuerySpecificationClassInferrer.this.inferBodies(pattern);
                    _builder.append(_inferBodies, "");
                    _builder.newLineIfNotEmpty();
                    StringConcatenationClient _inferAnnotations = PatternQuerySpecificationClassInferrer.this.inferAnnotations(pattern);
                    _builder.append(_inferAnnotations, "");
                    _builder.newLineIfNotEmpty();
                  } else {
                    _builder.append("try {");
                    _builder.newLine();
                    _builder.append("    ");
                    StringConcatenationClient _inferBodies_1 = PatternQuerySpecificationClassInferrer.this.inferBodies(pattern);
                    _builder.append(_inferBodies_1, "    ");
                    _builder.newLineIfNotEmpty();
                    _builder.append("    ");
                    StringConcatenationClient _inferAnnotations_1 = PatternQuerySpecificationClassInferrer.this.inferAnnotations(pattern);
                    _builder.append(_inferAnnotations_1, "    ");
                    _builder.newLineIfNotEmpty();
                    _builder.append("} catch (");
                    _builder.append(ViatraQueryException.class, "");
                    _builder.append(" ex) {");
                    _builder.newLineIfNotEmpty();
                    _builder.append("    ");
                    _builder.append("throw processDependencyException(ex);");
                    _builder.newLine();
                    _builder.append("}");
                    _builder.newLine();
                  }
                }
                _builder.append("return bodies;");
                _builder.newLine();
              }
            };
            PatternQuerySpecificationClassInferrer.this._eMFJvmTypesBuilder.setBody(it, _client);
          } catch (final Throwable _t) {
            if (_t instanceof Exception) {
              final Exception e = (Exception)_t;
              StringConcatenationClient _client_1 = new StringConcatenationClient() {
                @Override
                protected void appendTo(StringConcatenationClient.TargetStringConcatenation _builder) {
                  _builder.append("addError(new ");
                  _builder.append(PProblem.class, "");
                  _builder.append("(\"Inconsistent pattern definition threw exception ");
                  Class<? extends Exception> _class = e.getClass();
                  String _simpleName = _class.getSimpleName();
                  _builder.append(_simpleName, "");
                  _builder.append("  with message: ");
                  String _message = e.getMessage();
                  String _escapeToQuotedString = PatternQuerySpecificationClassInferrer.this.util.escapeToQuotedString(_message);
                  _builder.append(_escapeToQuotedString, "");
                  _builder.append("\"));");
                  _builder.newLineIfNotEmpty();
                  _builder.append("return bodies;");
                  _builder.newLine();
                }
              };
              PatternQuerySpecificationClassInferrer.this._eMFJvmTypesBuilder.setBody(it, _client_1);
            } else {
              throw Exceptions.sneakyThrow(_t);
            }
          }
        }
      };
      JvmOperation _method_3 = this._eMFJvmTypesBuilder.toMethod(pattern, "doGetContainedBodies", _typeRef_10, _function_6);
      _xblockexpression = this._eMFJvmTypesBuilder.<JvmOperation>operator_add(_members_6, _method_3);
    }
    return _xblockexpression;
  }
  
  public ExecutionType getRequestedExecutionType(final Pattern pattern) {
    ExecutionType _xblockexpression = null;
    {
      final Modifiers modifier = pattern.getModifiers();
      ExecutionType _xifexpression = null;
      if ((modifier != null)) {
        _xifexpression = modifier.getExecution();
      } else {
        _xifexpression = ExecutionType.UNSPECIFIED;
      }
      _xblockexpression = _xifexpression;
    }
    return _xblockexpression;
  }
  
  public StringConcatenationClient inferQueryEvaluationHints(final Pattern pattern) {
    StringConcatenationClient _switchResult = null;
    ExecutionType _requestedExecutionType = this.getRequestedExecutionType(pattern);
    if (_requestedExecutionType != null) {
      switch (_requestedExecutionType) {
        case INCREMENTAL:
          StringConcatenationClient _client = new StringConcatenationClient() {
            @Override
            protected void appendTo(StringConcatenationClient.TargetStringConcatenation _builder) {
              _builder.append("setEvaluationHints(new ");
              _builder.append(QueryEvaluationHint.class, "");
              _builder.append("(null, new ");
              _builder.append(ReteBackendFactory.class, "");
              _builder.append("()));");
            }
          };
          _switchResult = _client;
          break;
        case SEARCH:
          StringConcatenationClient _client_1 = new StringConcatenationClient() {
            @Override
            protected void appendTo(StringConcatenationClient.TargetStringConcatenation _builder) {
              _builder.append("setEvaluationHints(new ");
              _builder.append(QueryEvaluationHint.class, "");
              _builder.append("(null, ");
              _builder.append(LocalSearchBackendFactory.class, "");
              _builder.append(".INSTANCE»));");
            }
          };
          _switchResult = _client_1;
          break;
        case UNSPECIFIED:
          StringConcatenationClient _client_2 = new StringConcatenationClient() {
            @Override
            protected void appendTo(StringConcatenationClient.TargetStringConcatenation _builder) {
            }
          };
          _switchResult = _client_2;
          break;
        default:
          break;
      }
    }
    return _switchResult;
  }
  
  public StringConcatenationClient inferBodies(final Pattern pattern) throws IllegalStateException {
    StringConcatenationClient _client = new StringConcatenationClient() {
      @Override
      protected void appendTo(StringConcatenationClient.TargetStringConcatenation _builder) {
        {
          EList<PatternBody> _bodies = pattern.getBodies();
          for(final PatternBody body : _bodies) {
            _builder.newLineIfNotEmpty();
            _builder.append("{");
            _builder.newLine();
            _builder.append("    ");
            _builder.append("PBody body = new PBody(this);");
            _builder.newLine();
            _builder.append("    ");
            BodyCodeGenerator _bodyCodeGenerator = new BodyCodeGenerator(pattern, body, PatternQuerySpecificationClassInferrer.this.util, PatternQuerySpecificationClassInferrer.this.feedback, PatternQuerySpecificationClassInferrer.this.serializer, PatternQuerySpecificationClassInferrer.this.builder);
            _builder.append(_bodyCodeGenerator, "    ");
            _builder.newLineIfNotEmpty();
            _builder.append("    ");
            _builder.append("bodies.add(body);");
            _builder.newLine();
            _builder.append("}");
            _builder.newLine();
          }
        }
      }
    };
    return _client;
  }
  
  /**
   * Infers inner class for QuerySpecification class based on the input 'pattern'.
   */
  public boolean inferQuerySpecificationInnerClasses(final JvmDeclaredType querySpecificationClass, final Pattern pattern, final boolean withPatternSpecificMatcher) {
    boolean _xblockexpression = false;
    {
      EList<JvmMember> _members = querySpecificationClass.getMembers();
      String _querySpecificationHolderClassName = this.util.querySpecificationHolderClassName(pattern);
      final Procedure1<JvmGenericType> _function = new Procedure1<JvmGenericType>() {
        @Override
        public void apply(final JvmGenericType it) {
          it.setVisibility(JvmVisibility.PRIVATE);
          it.setStatic(true);
          StringConcatenation _builder = new StringConcatenation();
          _builder.append("Inner class allowing the singleton instance of {@link ");
          JvmType _findInferredSpecification = PatternQuerySpecificationClassInferrer.this.util.findInferredSpecification(pattern);
          _builder.append(_findInferredSpecification, "");
          _builder.append("} to be created ");
          _builder.newLineIfNotEmpty();
          _builder.append("    ");
          _builder.append("<b>not</b> at the class load time of the outer class, ");
          _builder.newLine();
          _builder.append("    ");
          _builder.append("but rather at the first call to {@link ");
          JvmType _findInferredSpecification_1 = PatternQuerySpecificationClassInferrer.this.util.findInferredSpecification(pattern);
          _builder.append(_findInferredSpecification_1, "    ");
          _builder.append("#instance()}.");
          _builder.newLineIfNotEmpty();
          _builder.newLine();
          _builder.append("<p> This workaround is required e.g. to support recursion.");
          _builder.newLine();
          PatternQuerySpecificationClassInferrer.this._eMFJvmTypesBuilder.setDocumentation(it, _builder.toString());
          EList<JvmMember> _members = it.getMembers();
          JvmTypeReference _typeRef = PatternQuerySpecificationClassInferrer.this.builder.typeRef(querySpecificationClass);
          final Procedure1<JvmField> _function = new Procedure1<JvmField>() {
            @Override
            public void apply(final JvmField it) {
              it.setFinal(true);
              it.setStatic(true);
              StringConcatenationClient _client = new StringConcatenationClient() {
                @Override
                protected void appendTo(StringConcatenationClient.TargetStringConcatenation _builder) {
                  _builder.append("new ");
                  JvmType _findInferredSpecification = PatternQuerySpecificationClassInferrer.this.util.findInferredSpecification(pattern);
                  _builder.append(_findInferredSpecification, "");
                  _builder.append("()");
                }
              };
              PatternQuerySpecificationClassInferrer.this._eMFJvmTypesBuilder.setInitializer(it, _client);
            }
          };
          JvmField _field = PatternQuerySpecificationClassInferrer.this._eMFJvmTypesBuilder.toField(pattern, "INSTANCE", _typeRef, _function);
          PatternQuerySpecificationClassInferrer.this._eMFJvmTypesBuilder.<JvmField>operator_add(_members, _field);
          EList<JvmMember> _members_1 = it.getMembers();
          JvmTypeReference _typeRef_1 = PatternQuerySpecificationClassInferrer.this.builder.typeRef(Object.class);
          final Procedure1<JvmField> _function_1 = new Procedure1<JvmField>() {
            @Override
            public void apply(final JvmField it) {
              it.setFinal(true);
              it.setStatic(true);
              StringConcatenationClient _client = new StringConcatenationClient() {
                @Override
                protected void appendTo(StringConcatenationClient.TargetStringConcatenation _builder) {
                  _builder.append("ensureInitialized()");
                }
              };
              PatternQuerySpecificationClassInferrer.this._eMFJvmTypesBuilder.setInitializer(it, _client);
              StringConcatenation _builder = new StringConcatenation();
              _builder.append("Statically initializes the query specification <b>after</b> the field {@link #INSTANCE} is assigned.");
              _builder.newLine();
              _builder.append("This initialization order is required to support indirect recursion.");
              _builder.newLine();
              _builder.newLine();
              _builder.append("<p> The static initializer is defined using a helper field to work around limitations of the code generator.");
              _builder.newLine();
              PatternQuerySpecificationClassInferrer.this._eMFJvmTypesBuilder.setDocumentation(it, _builder.toString());
            }
          };
          JvmField _field_1 = PatternQuerySpecificationClassInferrer.this._eMFJvmTypesBuilder.toField(pattern, "STATIC_INITIALIZER", _typeRef_1, _function_1);
          PatternQuerySpecificationClassInferrer.this._eMFJvmTypesBuilder.<JvmField>operator_add(_members_1, _field_1);
          EList<JvmMember> _members_2 = it.getMembers();
          JvmTypeReference _typeRef_2 = PatternQuerySpecificationClassInferrer.this.builder.typeRef(Object.class);
          final Procedure1<JvmOperation> _function_2 = new Procedure1<JvmOperation>() {
            @Override
            public void apply(final JvmOperation it) {
              it.setVisibility(JvmVisibility.PUBLIC);
              it.setStatic(true);
              StringConcatenationClient _client = new StringConcatenationClient() {
                @Override
                protected void appendTo(StringConcatenationClient.TargetStringConcatenation _builder) {
                  _builder.append("INSTANCE.ensureInitializedInternalSneaky();");
                  _builder.newLine();
                  _builder.append("return null;");
                  _builder.newLine();
                }
              };
              PatternQuerySpecificationClassInferrer.this._eMFJvmTypesBuilder.setBody(it, _client);
            }
          };
          JvmOperation _method = PatternQuerySpecificationClassInferrer.this._eMFJvmTypesBuilder.toMethod(pattern, "ensureInitialized", _typeRef_2, _function_2);
          PatternQuerySpecificationClassInferrer.this._eMFJvmTypesBuilder.<JvmOperation>operator_add(_members_2, _method);
        }
      };
      JvmGenericType _class = this._eMFJvmTypesBuilder.toClass(pattern, _querySpecificationHolderClassName, _function);
      this._eMFJvmTypesBuilder.<JvmGenericType>operator_add(_members, _class);
      EList<JvmMember> _members_1 = querySpecificationClass.getMembers();
      String _querySpecificationPQueryClassName = this.util.querySpecificationPQueryClassName(pattern);
      final Procedure1<JvmGenericType> _function_1 = new Procedure1<JvmGenericType>() {
        @Override
        public void apply(final JvmGenericType it) {
          it.setVisibility(JvmVisibility.PRIVATE);
          it.setStatic(true);
          EList<JvmTypeReference> _superTypes = it.getSuperTypes();
          JvmTypeReference _typeRef = PatternQuerySpecificationClassInferrer.this.builder.typeRef(BaseGeneratedEMFPQuery.class);
          PatternQuerySpecificationClassInferrer.this._eMFJvmTypesBuilder.<JvmTypeReference>operator_add(_superTypes, _typeRef);
          PatternQuerySpecificationClassInferrer.this.inferPQueryMembers(it, pattern);
        }
      };
      JvmGenericType _class_1 = this._eMFJvmTypesBuilder.toClass(pattern, _querySpecificationPQueryClassName, _function_1);
      _xblockexpression = this._eMFJvmTypesBuilder.<JvmGenericType>operator_add(_members_1, _class_1);
    }
    return _xblockexpression;
  }
  
  public void inferExpressions(final JvmDeclaredType querySpecificationClass, final Pattern pattern) {
    EList<PatternBody> _bodies = pattern.getBodies();
    final Function1<PatternBody, Collection<XExpression>> _function = new Function1<PatternBody, Collection<XExpression>>() {
      @Override
      public Collection<XExpression> apply(final PatternBody it) {
        return CorePatternLanguageHelper.getAllTopLevelXBaseExpressions(it);
      }
    };
    List<Collection<XExpression>> _map = ListExtensions.<PatternBody, Collection<XExpression>>map(_bodies, _function);
    Iterable<XExpression> _flatten = Iterables.<XExpression>concat(_map);
    final Procedure1<XExpression> _function_1 = new Procedure1<XExpression>() {
      @Override
      public void apply(final XExpression ex) {
        EList<JvmMember> _members = querySpecificationClass.getMembers();
        String _expressionMethodName = PatternQuerySpecificationClassInferrer.this.util.expressionMethodName(ex);
        JvmTypeReference _inferredType = PatternQuerySpecificationClassInferrer.this._eMFJvmTypesBuilder.inferredType(ex);
        final Procedure1<JvmOperation> _function = new Procedure1<JvmOperation>() {
          @Override
          public void apply(final JvmOperation it) {
            it.setVisibility(JvmVisibility.PRIVATE);
            it.setStatic(true);
            List<Variable> _variables = PatternQuerySpecificationClassInferrer.this.util.variables(ex);
            for (final Variable variable : _variables) {
              {
                String _name = variable.getName();
                JvmTypeReference _calculateType = PatternQuerySpecificationClassInferrer.this.util.calculateType(variable);
                final JvmFormalParameter parameter = PatternQuerySpecificationClassInferrer.this._eMFJvmTypesBuilder.toParameter(variable, _name, _calculateType);
                EList<JvmFormalParameter> _parameters = it.getParameters();
                PatternQuerySpecificationClassInferrer.this._eMFJvmTypesBuilder.<JvmFormalParameter>operator_add(_parameters, parameter);
              }
            }
            PatternQuerySpecificationClassInferrer.this._eMFJvmTypesBuilder.setBody(it, ex);
          }
        };
        JvmOperation _method = PatternQuerySpecificationClassInferrer.this._eMFJvmTypesBuilder.toMethod(ex, _expressionMethodName, _inferredType, _function);
        PatternQuerySpecificationClassInferrer.this._eMFJvmTypesBuilder.<JvmOperation>operator_add(_members, _method);
      }
    };
    IterableExtensions.<XExpression>forEach(_flatten, _function_1);
  }
  
  public StringConcatenationClient parameterInstantiation(final Variable variable) {
    StringConcatenationClient _xblockexpression = null;
    {
      final JvmTypeReference ref = this._iTypeInferrer.getJvmType(variable, variable);
      String _xifexpression = null;
      boolean _or = false;
      if ((ref == null)) {
        _or = true;
      } else {
        _or = (ref instanceof JvmUnknownTypeReference);
      }
      if (_or) {
        _xifexpression = "";
      } else {
        JvmType _type = ref.getType();
        _xifexpression = _type.getQualifiedName();
      }
      final String clazz = _xifexpression;
      final Type type = variable.getType();
      StringConcatenationClient _xifexpression_1 = null;
      boolean _or_1 = false;
      if ((type == null)) {
        _or_1 = true;
      } else {
        boolean _isValidType = this.typeSystem.isValidType(type);
        boolean _not = (!_isValidType);
        _or_1 = _not;
      }
      if (_or_1) {
        StringConcatenationClient _client = new StringConcatenationClient() {
          @Override
          protected void appendTo(StringConcatenationClient.TargetStringConcatenation _builder) {
            _builder.append("new PParameter(\"");
            String _name = variable.getName();
            _builder.append(_name, "");
            _builder.append("\", \"");
            _builder.append(clazz, "");
            _builder.append("\", (");
            _builder.append(IInputKey.class, "");
            _builder.append(")null, ");
            StringConcatenationClient _directionLiteral = PatternQuerySpecificationClassInferrer.this.directionLiteral(variable);
            _builder.append(_directionLiteral, "");
            _builder.append(")");
          }
        };
        _xifexpression_1 = _client;
      } else {
        StringConcatenationClient _xblockexpression_1 = null;
        {
          final IInputKey declaredInputKey = this.typeSystem.extractTypeDescriptor(type);
          StringConcatenationClient _client_1 = new StringConcatenationClient() {
            @Override
            protected void appendTo(StringConcatenationClient.TargetStringConcatenation _builder) {
              _builder.append("new PParameter(\"");
              String _name = variable.getName();
              _builder.append(_name, "");
              _builder.append("\", \"");
              _builder.append(clazz, "");
              _builder.append("\", ");
              StringConcatenationClient _serializeInputKey = PatternQuerySpecificationClassInferrer.this.util.serializeInputKey(declaredInputKey, true);
              _builder.append(_serializeInputKey, "");
              _builder.append(", ");
              StringConcatenationClient _directionLiteral = PatternQuerySpecificationClassInferrer.this.directionLiteral(variable);
              _builder.append(_directionLiteral, "");
              _builder.append(")");
            }
          };
          _xblockexpression_1 = _client_1;
        }
        _xifexpression_1 = _xblockexpression_1;
      }
      _xblockexpression = _xifexpression_1;
    }
    return _xblockexpression;
  }
  
  public StringConcatenationClient inferAnnotations(final Pattern pattern) {
    StringConcatenationClient _client = new StringConcatenationClient() {
      @Override
      protected void appendTo(StringConcatenationClient.TargetStringConcatenation _builder) {
        {
          EList<Annotation> _annotations = pattern.getAnnotations();
          for(final Annotation annotation : _annotations) {
            _builder.append("{");
            _builder.newLine();
            _builder.append("    ");
            _builder.append(PAnnotation.class, "    ");
            _builder.append(" annotation = new ");
            _builder.append(PAnnotation.class, "    ");
            _builder.append("(\"");
            String _name = annotation.getName();
            _builder.append(_name, "    ");
            _builder.append("\");");
            _builder.newLineIfNotEmpty();
            {
              LinkedHashMultimap<String, Object> _evaluateAnnotationParametersWithMultiplicity = CorePatternLanguageHelper.evaluateAnnotationParametersWithMultiplicity(annotation);
              Set<Map.Entry<String, Object>> _entries = _evaluateAnnotationParametersWithMultiplicity.entries();
              for(final Map.Entry<String, Object> attribute : _entries) {
                _builder.append("    ");
                _builder.append("annotation.addAttribute(\"");
                String _key = attribute.getKey();
                _builder.append(_key, "    ");
                _builder.append("\", ");
                Object _value = attribute.getValue();
                StringConcatenationClient _outputAnnotationParameter = PatternQuerySpecificationClassInferrer.this.outputAnnotationParameter(_value);
                _builder.append(_outputAnnotationParameter, "    ");
                _builder.append(");");
                _builder.newLineIfNotEmpty();
              }
            }
            _builder.append("    ");
            _builder.append("addAnnotation(annotation);");
            _builder.newLine();
            _builder.append("}");
            _builder.newLine();
          }
        }
      }
    };
    return _client;
  }
  
  public StringConcatenationClient outputAnnotationParameter(final Object value) {
    StringConcatenationClient _switchResult = null;
    boolean _matched = false;
    if (!_matched) {
      if (value instanceof List) {
        _matched=true;
        StringConcatenationClient _client = new StringConcatenationClient() {
          @Override
          protected void appendTo(StringConcatenationClient.TargetStringConcatenation _builder) {
            _builder.append(Arrays.class, "");
            _builder.append(".asList(new Object[] {");
            _builder.newLineIfNotEmpty();
            {
              boolean _hasElements = false;
              for(final Object item : ((List<?>)value)) {
                if (!_hasElements) {
                  _hasElements = true;
                } else {
                  _builder.appendImmediate(", ", "                    ");
                }
                _builder.append("                    ");
                StringConcatenationClient _outputAnnotationParameter = PatternQuerySpecificationClassInferrer.this.outputAnnotationParameter(item);
                _builder.append(_outputAnnotationParameter, "                    ");
                _builder.newLineIfNotEmpty();
              }
            }
            _builder.append("                    ");
            _builder.append("})");
          }
        };
        _switchResult = _client;
      }
    }
    if (!_matched) {
      if (value instanceof ParameterReference) {
        _matched=true;
        StringConcatenationClient _client = new StringConcatenationClient() {
          @Override
          protected void appendTo(StringConcatenationClient.TargetStringConcatenation _builder) {
            _builder.append("new ");
            _builder.append(ParameterReference.class, "");
            _builder.append("(\"");
            String _name = ((ParameterReference)value).getName();
            _builder.append(_name, "");
            _builder.append("\")");
          }
        };
        _switchResult = _client;
      }
    }
    if (!_matched) {
      if (value instanceof String) {
        _matched=true;
        StringConcatenationClient _client = new StringConcatenationClient() {
          @Override
          protected void appendTo(StringConcatenationClient.TargetStringConcatenation _builder) {
            _builder.append("\"");
            _builder.append(((String)value), "");
            _builder.append("\"");
          }
        };
        _switchResult = _client;
      }
    }
    if (!_matched) {
      StringConcatenationClient _client = new StringConcatenationClient() {
        @Override
        protected void appendTo(StringConcatenationClient.TargetStringConcatenation _builder) {
          String _string = value.toString();
          _builder.append(_string, "");
        }
      };
      _switchResult = _client;
    }
    return _switchResult;
  }
}
